﻿using AliGenieAssistant;
using MQTTnet;
using MQTTnet.Client;
using System;
using System.Collections.ObjectModel;
using System.Text.Json;
using System.Threading;
using System.Threading.Tasks;
using Windows.Storage;

namespace AliGenieAssistantWinUI3
{
    class MQTTHelper
    {
        private static MQTTHelper _instance;

        private IMqttClient client;

        MqttClientOptions mqttOption;

        AppHelper helper;

        Windows.Storage.StorageFolder storageFolder = Windows.Storage.ApplicationData.Current.LocalFolder;

        String configFileName = "config.json";
        private string identify;

        public String ip { set; get; } = "www.xiongweixp.tech";

        public int port { set; get; } = 1883;
        private MQTTHelper() {

            helper = AppHelper.getInstance();
            MqttFactory mqttFactory = new MqttFactory();
            client = mqttFactory.CreateMqttClient();
            client.DisconnectedAsync += Client_DisconnectedAsync;
            client.ConnectedAsync += Client_ConnectedAsync;
            client.ApplicationMessageReceivedAsync += Client_ApplicationMessageReceivedAsync;
        }

        public static MQTTHelper Instance()
        {
            if (_instance == null) {
                _instance = new MQTTHelper();
            }
            return _instance;
        }

        public async void Start()
        {
            await loadConfig();
            initMQTT();
        }

        private async Task<bool> loadConfig()
        {
           
            if (await s_doesFileExistAsync( storageFolder, configFileName))
            {
                Windows.Storage.StorageFile configFile =
                        await storageFolder.GetFileAsync(configFileName);
                string configText = await Windows.Storage.FileIO.ReadTextAsync(configFile);
                Url configUrl = JsonSerializer.Deserialize<Url>(configText);
                if ( !configUrl.ip.Equals("") && configUrl.port!=0)
                {
                    ip = configUrl.ip;
                    port = configUrl.port;
                    return true;
                }

            }
            else
            {    
                Windows.Storage.StorageFile configFile =
                        await storageFolder.CreateFileAsync(configFileName,
                            Windows.Storage.CreationCollisionOption.OpenIfExists);
 
                Url configUrl = new Url();
                configUrl.ip = ip;
                configUrl.port = port;
                String configText = JsonSerializer.Serialize(configUrl);
                var buffer = Windows.Security.Cryptography.CryptographicBuffer.ConvertStringToBinary(
                       configText, Windows.Security.Cryptography.BinaryStringEncoding.Utf8);
                await Windows.Storage.FileIO.WriteBufferAsync(configFile, buffer);
            }
            return true;
        }
        static async private Task<bool> s_doesFileExistAsync(StorageFolder storageFolder, string fileName)
        {
            try
            {
                await storageFolder.GetFileAsync(fileName);
                return true;
            }
            catch
            {
                return false;
            }
        }

        private async void initMQTT()
        {
            mqttOption = createMqttOptions();
            await client.ConnectAsync(mqttOption, CancellationToken.None);

        }
        private async Task Client_ConnectedAsync(MqttClientConnectedEventArgs arg)
        {

            addLog("连接服务器"+ip+":"+port+"成功");
            string topic = "$share/pc-" + identify + "/pccontrol";
            await client.SubscribeAsync(new MqttTopicFilterBuilder().WithTopic(topic).Build());
            addLog("订阅主题:"+topic);

        }

        private async Task Client_DisconnectedAsync(MqttClientDisconnectedEventArgs arg)
        {
            addLog("连接断开，5秒后尝试重新连接");
            await Task.Delay(TimeSpan.FromSeconds(5));
            try
            {
                mqttOption = createMqttOptions();
                await client.ConnectAsync(mqttOption, CancellationToken.None);
            }
            catch
            {
                addLog("重新连接失败");
                await Task.Delay(TimeSpan.FromSeconds(5));
            }
        }

        private MqttClientOptions createMqttOptions()
        {
            Guid g = Guid.NewGuid();
            string[] identifies = g.ToString().Split("-");
            identify = identifies[identifies.Length-1];
            identify = identify.Substring(identify.Length - 4);
            return new MqttClientOptionsBuilder()
                    .WithTcpServer(ip, port)
                    .WithClientId("pcassistant-"+identify)
                    .WithProtocolVersion(MQTTnet.Formatter.MqttProtocolVersion.V500)
                    .Build();
        }

        public async Task Reset()
        {
            addLog("准备断开连接");
            await client.DisconnectAsync();
            addLog("连接已经断开");

        }

        private Task Client_ApplicationMessageReceivedAsync(MqttApplicationMessageReceivedEventArgs arg)
        {
            var payload = System.Text.Encoding.UTF8.GetString(arg.ApplicationMessage.Payload);
            addLog("收到消息,payload:" + payload);
            //  解析JSON 
            Operator opt = JsonSerializer.Deserialize<Operator>(payload);
            addLog(opt.intentName + ":" + opt.slotValue);
            helper.controlApp(opt);
            return Task.CompletedTask;
        }

        private void addLog(String logMsg)
        {
            LoggerHelper.Instance().AddLog(logMsg);
        }

        public async void ChangeConfig(String new_ip, int new_port)
        {
            if (new_ip.Equals("") || new_port == 0)
            {
                return;
            }
            Windows.Storage.StorageFile configFile =
                        await storageFolder.CreateFileAsync(configFileName,
                            Windows.Storage.CreationCollisionOption.ReplaceExisting);

            Url configUrl = new Url();
            configUrl.ip = new_ip;
            configUrl.port = new_port;
            String configText = JsonSerializer.Serialize(configUrl);
            var buffer = Windows.Security.Cryptography.CryptographicBuffer.ConvertStringToBinary(
                   configText, Windows.Security.Cryptography.BinaryStringEncoding.Utf8);
            await Windows.Storage.FileIO.WriteBufferAsync(configFile, buffer);
            ip= new_ip;
            port= new_port;
        }
    }

    internal class Url
    {
        public String? ip { set; get; }
        public int port { set; get; }
    }

}
